#ifndef _LOAD3DS_H_
#define _LOAD3DS_H_

/*
  zjednoduseny popis obsahu 3ds suboru:

  CHUNK       - informacny blok (infoblok) 3ds suboru, v ktorom su ulozene udaje,
                typ infobloku urcuje, ake udaje sa v nom nachadzaju
  CHUNKLENGTH - infoblok nema informaciu o velkosti svojich udajov ani pocte podblokov,
                ale ma informaciu o svojej celkovej velkosti
  DATA        - infoblok moze/nemusi obsahovat urcity druh udajov
  SUBCHUNKS   - infoblok moze/nemusi obsahovat infoblok, ktory moze obsahovat dalsi infoblok,
                v infobloku moze byt takymto sposobom (rekurzivne) umiestnenych viacero infoblokov
  OBJECT      - suradnice vrcholov, popis stien sietovych modelov a dalsie vlastnosti
                definovanych objektov
  MATERIAL    - farba, alebo textura priradena stenam objektu, material navyse uchovava
                informacie o optickych vlastnostiach povrchu stien

  z 3dsinfo.txt:

  This is the representation of a chunk :

  Offset   Length    Name
  0        2         Chunk-ID
  2        4         Chunk-length = 2+4+n+m
  6        n         Data
  6+n      m         Sub-chunks

  Type            Name                         Bytes read
  0x0011          RGB color
  0x4D4D          Main chunk                   2+4+   (id+length)
   0x0002          3DS-Version                 2+4+4+ (id+length+version)
   0x3D3D          3D editor chunk             2+4+   (id+length)
    0x4000          Object block               ...
     0x4100          Triangular mesh
       0x4110         Vertices lists
       0x4120         Faces description
        0x4130         Faces material list
       0x4140         Mapping coordinates list
    0xAFFF          Material block
    0xA000          Material name
     0xA010          Ambient color
     0xA020          Diffuse color
     0xA030          Specular color
     0xA200          Texture map 1
      0xA300          Mapping filename
*/

// typy infoblokov 3ds suboru
#define RGB_COLOR      0x0011 // farba

#define MAIN           0x4D4D // hlavny infoblok
#define VERSION        0x0002 // verzia
#define EDITOR         0x3D3D // 3D editor

#define OBJECT         0x4000 // objekt
#define MESH           0x4100 // sietovy model objektu
#define VERTS_LIST     0x4110 // vrcholy sietoveho modelu
#define FACESDESCR     0x4120 // steny sietoveho modelu
#define FACESMAT_LIST  0x4130 // material stien
#define TEXCOORDS_LIST 0x4140 // texturovacie suradnice

#define MATERIAL       0xAFFF // material
#define MAT_NAME       0xA000 // nazov materialu
#define MAT_AMBIENT    0xA010 // ambientna zlozka farby
#define MAT_DIFFUSE    0xA020 // difuzna zlozka farby
#define MAT_SPECULAR   0xA030 // specular zlozka farby
#define MAT_MAP        0xA200 // textura
#define MAT_FILENAME   0xA300 // nazov suboru s texturou

// indexy vrcholov umiestnene do pola stien
struct Sindices {
 // 3x index vrcholu, 1x face flag (visibility)
 unsigned short int ind[4]; // 2 Byte x 4
};

// objekt
struct Sobject {
 char name[255];                  // nazov objektu - 1 Byte / prvok
 unsigned short int vertsNum;     // pocet vrcholov - 2 Byte
 vec3 *verts;                     // pole vrcholov - float x 3
 vec3 *normals;                   // pole normal
 unsigned short int facesNum;     // pocet stien - 2 Byte
 Sindices *faces;                 // pole stien - 2 Byte x 4
 int matIndex;                    // index do zoznamu materialov
 unsigned short int texcoordsNum; // pocet texturovacich suradnic - 2 Byte
 vec2 *texCoords;                 // pole texturovacich suradnic - float x 2
};

// material
struct Smaterial {
 char name[255];            // nazov materialu - 1 Byte / prvok
 unsigned char difColor[3]; // difuzna zlozka farby (R,G,B) - 1 Byte x 3
 char mapFile[255];         // nazov suboru s texturou - 1 Byte / prvok
 unsigned int mapTid;       // ciselne oznacenie textury 2-(MAX_TEX_NUM-1)
 unsigned int nmapTid;      // ciselne oznacenie normalmapy 2-(MAX_TEX_NUM-1)
};

// 3ds model
struct Smodel {
 unsigned int objNum;   // pocet objektov
 vector<Sobject> obj;   // zoznam objektov
 unsigned int matNum;   // pocet materialov
 vector<Smaterial> mat; // zoznam materialov
};

// infoblok 3ds suboru
struct Schunk {
 unsigned short int id; // Chunk-ID     - 2 Byte = 16 bit
 unsigned int length;   // Chunk-length - 4 Byte = 32 bit
 unsigned int Bread;    // pocet nacitanych Bytov
};

// trieda premennych a funkcii (objektov) pre nacitanie 3ds suborov
class CLoad3ds {
public:
 CLoad3ds();  // konstruktor
 ~CLoad3ds(); // destruktor

 int load(const char *fname); // nacitanie obsahu 3ds suboru
 int forcedraw(void);         // vykreslenie 3ds modelu

 // ziskanie trojuholnikov patriacich objektom
 // ziskanie informacii o svetelnych zdrojoch definovanych v 3ds subore
 int get_tris_and_lights(void);
private:
 FILE *fp; // pre pracu so suborom

 Smodel model; // 3ds model

 Sobject *tmpObj;   // pomocna premenna pre pristup k objektu
 Smaterial *tmpMat; // pomocna premenna pre pristup k materialu

 Schunk ch; // pomocna premenna pre nacitanie obsahu infoblokov

 unsigned int i,objn,matn,l;  // pomocne premenne

 unsigned int *buf; // pomocna pamat - 4 Byte / prvok

 unsigned int ver; // pomocna premenna - 4 Byte

 char tmpName[255]; // pomocna premenna - 1 Byte / prvok

 bool isTex; // indikator existencie textury

 // nacitanie Chunk-ID a Chunk-length infobloku
 void ReadChunkID(Schunk *C);
 // nacitanie obsahu infoblokov - REKURZIVNA FUNKCIA!
 void ReadChunk(Schunk *C);
 // nacitanie infobloku OBJECT - REKURZIVNA FUNKCIA!
 void ReadObjectChunk(Schunk *C);
 // nacitanie infobloku MATERIAL
 void ReadMaterialChunk(Schunk *C);

 // vypocet normal
 int ComputeNormals(void);
};

#endif
