#ifndef _RBCOLLISIONS_H_
#define _RBCOLLISIONS_H_

/*
 Pouzity sposob riesenia kolizii:

 1. Poloha kolizneho bodu je relativna voci stredu (tazisku) RB.
    Kolizna normala a tangenty su ulozene v matici col_n_t.
    Relativna rychlost RB pred aplikaciou impulzu je v0.

 2. Impulz sposobuje zmenu priamociarej a uhlovej rychlosti tak, aby sme po
    jeho aplikacii ziskali pozadovanu relativnu rychlost RB v1.

 3. Na zaklade znamej hodnoty v0 a pozadovanej hodnoty v1 je vypocitany impulz J,
    ktoreho aplikaciou ziskame pozadovanu zmenu rychlosti.

 4. Impulz J je rozdeleny na priamociaru a uhlovu zlozku, ktora sa aplikuje
    na obe RB v kolizii.
*/

#include "RB.h"

// maximalny mozny pocet detekovanych kolizii
#define MAX_COLLISIONS 256

// trieda pre spravu parametrov koliznej roviny ****************************
class CCollisionPlane
{
 public:
  // rovnica roviny Nx * X + Ny * Y + Nz * Z + D = 0
  // normala - prvky Nx, Ny, Nz rovnice roviny
  vec3 normal;
  // vzdialenost roviny od zaciatku suradnicovej sustavy - prvok D rovnice roviny
  float dist;
};

// trieda pre spravu parametrov kolizneho telesa - kvadra ******************
class CCollisionBox
{
 public:
  // kazde kolizne teleso ma pevne teleso
  CRigidBody *rb;

  // konstruktor
  CCollisionBox() { rb=new CRigidBody(); }
  // destruktor
  ~CCollisionBox() { DELETEP(rb); }
  // polovica rozmerov kvadra
  vec3 halfSize;

  // pomocna matica pre OpenGL transformacie
  float mat[16];
  // ukazovatel na trojuholnik
  triangle *t;
};

// parametre detekovanej kolizie *******************************************
class CCollision
{
 private:
  // prirastok rychlosti zavisly od zrychlenia posobiaceho
  // v danom casovom intervale delta t
  vec3 dv;
  // pomocna premenna
  float d;
  // pomocne vektory
  vec3 h0,h1;
  // pomocne matice pre vypocet j
  mat3x3 ssm,mat0,mat1,jm;
 public:
  // ukazovatel na RBa a RBb
  CRigidBody *RBa,*RBb;
  // poloha kolizneho bodu vo world space
  vec3 col_p;
  // normala v koliznom bode
  vec3 col_n;
  // tangenty v koliznom bode
  vec3 col_t[2];
  // hlbka vzajomneho prieniku RBa do RBb
  float col_penetr;
  // koeficient obnovy
  float cor;

  // matica pre ulozenie col_n, col_t
  mat3x3 col_n_t;
  // vzdialenost kolizneho bodu od stredu RBa a RBb
  vec3 r[2];
  // relativna rychlost RB pred aplikovanim impulzu
  vec3 v0;
  // delenec (numerator) a delitel (denominator) pre vypocet j
  float numer,denom;

  // vypocet col_t, col_n_t, r, dv, v0, numer
  void calc_params(float DT);

  // vypocet delenca (numerator) pre vypocet j - zavisly na v0
  void calc_numer(float DT);

  // vypocet velkosti impulzu so zanedbanim trenia
  vec3 calc_j_frictionless(mat3x3 Iia,mat3x3 Iib);
  // vypocet velkosti impulzu s aplikaciou trenia
  vec3 calc_j_friction(mat3x3 Iia,mat3x3 Iib);
};

// detekcia a riesenie kolizii *********************************************
class CCollisionDetection
{
 private:
  // pole obsahujuce parametre detekovanych kolizii
  CCollision col[MAX_COLLISIONS];
  // ukazovatel na 1. prvok a aktualny prvok pola, do ktoreho sa ulozia parametre
  // detekovanej kolizie
  CCollision *col_first,*col_act,*c;
  // aktualny pocet detekovanych kolizii a zostavajuci mozny pocet detekovanych kolizii
  unsigned int col_num,col_remain;

  // pomocne premenne
  float d,p_min;
  unsigned int k,i_min;
  // pomocne vektory
  vec3 pv,centerv,n;
  // RB, pre ktore je detekovana kolizia
  CCollisionBox b1,b2;

  // pocet opakovani riesenia detekovanych kolizii - poloha a rychlost
  unsigned int s_iter_num,v_iter_num,iter_used;

  // minimalna zmena polohy a rychlosti, pri ktorej budu riesene detekovane kolizie
  float s_epsilon,v_epsilon;

  // pomocne premenne
  unsigned int id;
  float max;

  // pomocne vektory
  vec3 d_v[2],d_w[2],d_vw;
  vec3 d_s[2],d_r[2],d_sr;

  // pomocne matice pre ulozenie I_inv
  mat3x3 Ii[2];
  // pomocne vektory
  vec3 h0,h1;
  // pomocne premenne
  float u[2],denom,sl[2],sa[2],sla;

  // impulz
  vec3 j,J,torqueJ;

  // reset pola parametrov detekovanych kolizii
  void col_reset(void) {
   // ukazovatel na 1. prvok pola
   col_first=col;
   // ukazovatel na aktualny prvok, do ktoreho sa ulozia parametre detekovanej kolizie
   col_act=col_first;
   col_num=0;                 // aktualny pocet detekovanych kolizii
   col_remain=MAX_COLLISIONS; // mozno detekovat MAX_COLLISIONS kolizii
  }

  // pridanie parametrov detekovanej kolizie do pola
  void add_collision(CRigidBody *B1,CRigidBody *B2,
                     vec3 CP,vec3 CN,float P,float C) {
   col_act->RBa=B1; col_act->RBb=B2;
   col_act->col_p=CP; col_act->col_n=CN; col_act->col_penetr=P;
   col_act->cor=C;

   col_num++;    // zvysenie aktualneho poctu detekovanych kolizii
   col_remain--; // znizenie poctu kolizii, ktore este mozno detekovat
   // ukazovatel na aktualny prvok pola, do ktoreho sa ulozia parametre detekovanej kolizie
   col_act++;
  }

  // vzdialenost stredu RB od jeho povrchu v smere vektora N
  float center_to_surface(const CCollisionBox &B,const vec3 N) {
   return(B.halfSize.x*fabs(N.dot(B.rb->get_rotX()))+
          B.halfSize.y*fabs(N.dot(B.rb->get_rotY()))+
          B.halfSize.z*fabs(N.dot(B.rb->get_rotZ())));
  }

  // detekcia kolizii medzi RB a koliznymi rovinami
  bool box_plane(const CCollisionBox &B,const CCollisionPlane &P);
  // test vzdialenosti stredu b2 (b1) od roviny prechadzajucej stredom b1 (b2)
  bool check_plane(vec3 N,const unsigned int I);
  // detekcia kolizii medzi RB a ostatnymi RB
  bool box_box(const CCollisionBox &B1,const CCollisionBox &B2);

  // zmena poloh a rotacnych matic RBa a RBb aplikovanim impulzu
  void change_s_R(CCollision *C,vec3 Ds[2],vec3 DR[2]);

  // zmena rychlosti RBa a RBb aplikovanim impulzu
  void change_v_w(CCollision *C,vec3 Dv[2],vec3 Dw[2]);

  // ak je jedno RB v kolizii aktivne a druhe neaktivne, neaktive sa aktivuje
  void activate_inactive_RB(void);
 public:
  // konstruktory
  CCollisionDetection() {
   set_iter_num(64); set_epsilon(0.01f,0.01f);
  }
  CCollisionDetection(unsigned int I,float A,float B) {
   set_iter_num(I); set_epsilon(A,B);
  }
  CCollisionDetection(unsigned int I1,unsigned int I2,float A,float B) {
   set_iter_num(I1,I2); set_epsilon(A,B);
  }

  // pocet opakovani riesenia detekovanych kolizii
  void set_iter_num(const unsigned int A) { set_iter_num(A,A); }
  void set_iter_num(const unsigned int A,const unsigned int B) {
   v_iter_num=A; s_iter_num=B;
  }

  // minimalna zmena polohy a rychlosti, pri ktorej budu riesene detekovane kolizie
  void set_epsilon(const float A,const float B) { v_epsilon=A; s_epsilon=B; }

  // detekcia a riesenie kolizii
  void detect_and_resolve_collisions(const CCollisionBox *B,const unsigned int BN,
                                     const CCollisionPlane *P,const unsigned int PN,
                                     float DT);
};

#endif
